/*
 * Copyright (C) 2005-2006 Andre Noll <maan@systemlinux.org>
 *
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 */

#include "para.h"
#include <sys/utsname.h> /* uname() */
#include "gcc-compat.h"
#include <netdb.h>

#include "error.h"

extern void (*crypt_function_recv)(unsigned long len, const unsigned char *indata, unsigned char *outdata);
extern void (*crypt_function_send)(unsigned long len, const unsigned char *indata, unsigned char *outdata);

/*
 * Init sockaddr. If he is null (server mode), use own IP, else use
 * that given in he.
 */
void init_sockaddr(struct sockaddr_in *addr, int port, struct hostent *he)
{
	/* host byte order */
	addr->sin_family = AF_INET;
	/* short, network byte order */
	addr->sin_port = htons(port);
	if (he)
		addr->sin_addr = *((struct in_addr *)he->h_addr);
	else
		addr->sin_addr.s_addr = INADDR_ANY;
	/* zero the rest of the struct */
	memset(&addr->sin_zero, '\0', 8);
}

/*
 * Taken from Beej's Guide to Network Programming
 *
 * Due to circumstances beyond your control, the kernel might not send
 * all the data out in one chunk, and now, my friend, it's up to us to
 * get the data out there.
 */
static int sendall(int fd, const char *buf, size_t *len)
{
	int total = 0;        /* how many bytes we've sent */
	int bytesleft = *len; /* how many we have left to send */
	int n = -1;

	while (total < *len) {
		n = send(fd, buf + total, bytesleft, 0);
		if (n == -1)
			break;
		total += n;
		bytesleft -= n;
		if (total < *len)
			PARA_DEBUG_LOG("short write (%d byte(s) left)",
				*len - total);
	}
	*len = total; /* return number actually sent here */
	return n == -1? -E_SEND : 1; /* return 1 on success */
}

/**
 * send_bin_buffer -- Send buffer of length len to fd.
 * Retry on short writes.
 */
int send_bin_buffer(int fd, const char *buf, size_t len)
{
	int ret;

	if (!len)
		PARA_CRIT_LOG("%s", "len == 0\n");
	if (crypt_function_send) {
		unsigned char *outbuf = para_malloc(len);
		crypt_function_send(len, (unsigned char *)buf, outbuf);
		ret = sendall(fd, (char *)outbuf, &len);
		free(outbuf);
	} else
		ret = sendall(fd, buf, &len);
	return ret;
}

/*
 * Send null terminated buffer to socket.
 */
int send_buffer(int socket_fd, char *buf)
{
	return send_bin_buffer(socket_fd, buf, strlen(buf));
}


__printf(2, 3) int send_va_buffer(int fd, char *fmt, ...) 
{
	char *msg;
	int ret;

	PARA_VSPRINTF(fmt, msg);
	ret = send_buffer(fd, msg);
	free(msg);
	return ret;
}

__must_check int recv_bin_buffer(int socket_fd, char *buf, ssize_t len)
{
	int n;

	if (crypt_function_recv) {
		unsigned char *tmp = para_malloc(len);
		n = recv(socket_fd, tmp, len, 0);
		if (n > 0)
			crypt_function_recv(n, tmp, (unsigned char *)buf);
		free(tmp);
	} else
		n = recv(socket_fd, buf, len, 0);
	if (n == -1)
		n = -E_RECV;
	return n;
}

int recv_buffer(int socket_fd, char *buf, ssize_t len)
{
        int n;

	if ((n = recv_bin_buffer(socket_fd, buf, len - 1)) >= 0)
		buf[n] = '\0';
	return n;
}

/* FIXME: gethostbyname() is obsolete */
struct hostent *get_host_info(char *host)
{
	struct hostent *he;

	if (!(he = gethostbyname(host)))
		PARA_ERROR_LOG("gethostbyname error (%s)\n", strerror(errno));
	return he;
}

int get_socket(void)
{
	int socket_fd;

	if ((socket_fd = socket(AF_INET, SOCK_STREAM, 0)) == -1)
		return -E_SOCKET;
	return socket_fd;
}

int do_connect(int socket_fd, struct sockaddr_in *their_addr)
{
	int ret;

	if ((ret = connect(socket_fd, (struct sockaddr *)their_addr,
			sizeof(struct sockaddr))) == -1)
		return -E_CONNECT;
	return ret;
}

/**
 * para_accept - simple wrapper around the accept system call
 * @socket_fd: The listening socket
 * @addr: structure which is filled in with the address of the peer socket
 * @size: should contain the size of the structure pointed to by @addr
 *
 * See accept(2).
 */
int para_accept(int socket_fd, void *addr, size_t size)
{
	int new_fd;

	new_fd = accept(socket_fd, (struct sockaddr *) addr, &size);
	return new_fd == -1? -E_ACCEPT : new_fd;
}

static int setserversockopts(int socket_fd)
{
	int yes = 1;

	if (setsockopt(socket_fd, SOL_SOCKET, SO_REUSEADDR, &yes,
			sizeof(int)) == -1)
		return -E_SETSOCKOPT;
	return 1;
}
static int do_bind(int socket_fd, struct sockaddr_in *my_addr)
{
	if (bind(socket_fd, (struct sockaddr *)my_addr,
		sizeof(struct sockaddr)) == -1)
		return -E_BIND;
	return 1;
}

int init_unix_addr(struct sockaddr_un *u, const char *name)
{
	if (strlen(name) >= UNIX_PATH_MAX)
		return -E_NAME_TOO_LONG;
	memset(u->sun_path, 0, UNIX_PATH_MAX);
	u->sun_family = PF_UNIX;
	strcpy(u->sun_path, name);
	return 1;
}

int create_pf_socket(const char *name, struct sockaddr_un *unix_addr, int mode)
{
	int fd, ret;

	fd = socket(PF_UNIX, SOCK_STREAM, 0);
	if (fd < 0)
		return -E_SOCKET;
//	unlink(name);
	ret = init_unix_addr(unix_addr, name);
	if (ret < 0)
		return ret;
	if (bind(fd, (struct sockaddr *) unix_addr, UNIX_PATH_MAX) < 0)
		return -E_BIND;
	if (chmod(name, mode) < 0)
		return -E_CHMOD;
	return fd;
}

ssize_t send_cred_buffer(int sock, char *buf)
{
	char control[sizeof(struct cmsghdr) + 10];
	struct msghdr msg;
	struct cmsghdr *cmsg;
	static struct iovec iov;
	struct ucred c;
	int ret;

	/* Response data */
	iov.iov_base = buf;
	iov.iov_len  = strlen(buf);
	c.pid = getpid();
	c.uid = getuid();
	c.gid = getgid();
	/* compose the message */
	memset(&msg, 0, sizeof(msg));
	msg.msg_iov = &iov;
	msg.msg_iovlen = 1;
	msg.msg_control = control;
	msg.msg_controllen = sizeof(control);
	/* attach open tty fd */
	cmsg = CMSG_FIRSTHDR(&msg);
	cmsg->cmsg_level = SOL_SOCKET;
	cmsg->cmsg_type = SCM_CREDENTIALS;
	cmsg->cmsg_len = CMSG_LEN(sizeof(struct ucred));
	*(struct ucred *)CMSG_DATA(cmsg) = c;
	msg.msg_controllen = cmsg->cmsg_len;
	ret = sendmsg(sock, &msg, 0);
	if (ret  < 0)
		ret = -E_SENDMSG;
	return ret;
}

static void dispose_fds(int *fds, int num)
{
	int i;

	for (i = 0; i < num; i++)
		close(fds[i]);
}

/*
 * taken from Olaf Kirch's Black Hats Manual
 */
int recv_cred_buffer(int fd, char *buffer, size_t size, struct ucred *cred)
{
	char control[255];
	struct msghdr msg;
	struct cmsghdr *cmsg;
	struct iovec iov;
	int result;
	int yes = 1;

	setsockopt(fd, SOL_SOCKET, SO_PASSCRED, &yes, sizeof(int));
	memset(&msg, 0, sizeof(msg));
	memset(buffer, 0, size);
	iov.iov_base = buffer;
	iov.iov_len = size;
	msg.msg_iov = &iov;
	msg.msg_iovlen = 1;
	msg.msg_control = control;
	msg.msg_controllen = sizeof(control);
	if (recvmsg(fd, &msg, 0) < 0)
		return -E_RECVMSG;
	result = -SCM_CREDENTIALS;
	cmsg = CMSG_FIRSTHDR(&msg);
	while (cmsg) {
		if (cmsg->cmsg_level == SOL_SOCKET
		 && cmsg->cmsg_type == SCM_CREDENTIALS) {
			memcpy(cred, CMSG_DATA(cmsg), sizeof(struct ucred));
			result = iov.iov_len;
		} else
		if (cmsg->cmsg_level == SOL_SOCKET
				&& cmsg->cmsg_type == SCM_RIGHTS) {
			dispose_fds((int *) CMSG_DATA(cmsg),
				(cmsg->cmsg_len - CMSG_LEN(0))/sizeof(int));
		}
		cmsg = CMSG_NXTHDR(&msg, cmsg);
	}
	return result;
}

char *para_hostname(void)
{
	struct utsname u;

	uname(&u);
	return para_strdup(u.nodename);
}

#define BACKLOG 10 /*  how many pending connections queue will hold */
int init_tcp_socket(int port)
{
	int sockfd, ret;
	struct sockaddr_in my_addr;

	if ((sockfd = get_socket()) < 0)
		return sockfd;
	ret = setserversockopts(sockfd); 
	if (ret < 0)
		return ret;
	init_sockaddr(&my_addr, port, NULL);
	ret = do_bind(sockfd, &my_addr);
	if (ret < 0)
		return ret;
	if (listen(sockfd, BACKLOG) == -1)
		return -E_LISTEN;
	PARA_INFO_LOG("listening on port %d, fd %d\n", port, sockfd);
	return sockfd;
}

int recv_pattern(int fd, const char *pattern, size_t bufsize)
{
	size_t len = strlen(pattern);
	char *buf = para_malloc(bufsize + 1);
	int ret = -E_RECV_PATTERN, n = recv_buffer(fd, buf, bufsize);

	if (n < len)
		goto out;
	buf[n] = '\0';
	if (strncasecmp(buf, pattern, len))
		goto out;
	ret = 1;
out:
	free(buf);
	return ret;
}
